use crate::{client::async_http_client, config::Config, db::DbThread, models::Account};
use actix_session::Session;
use actix_web::{
    get,
    http::header::{ContentType, LOCATION},
    web::{self, ServiceConfig},
    HttpResponse, Responder,
};
use eyre::{ensure, OptionExt};
use openidconnect::{
    core::{CoreAuthenticationFlow, CoreUserInfoClaims},
    AuthorizationCode, CsrfToken, EndUserEmail, Nonce, OAuth2TokenResponse, TokenResponse,
};
use serde::{Deserialize, Serialize};
use std::marker::PhantomData;
use tinytemplate::TinyTemplate;

pub trait DynTemplate {
    fn name(&self) -> &'static str;
    fn text(&self) -> &'static str;
}

pub trait Template: Serialize
where
    PhantomData<Self>: DynTemplate,
{
    const NAME: &'static str;
    const TEXT: &'static str;
    fn render(&self) -> tinytemplate::error::Result<String> {
        with_templates(|templates| templates.render(Self::NAME, &self))
    }
}

impl<T: Template> DynTemplate for PhantomData<T> {
    fn name(&self) -> &'static str {
        T::NAME
    }

    fn text(&self) -> &'static str {
        T::TEXT
    }
}

macro_rules! templates {
    (
        $(
            #[text = $text:literal]
            $(#[$meta:meta])*
            $vis:vis struct $name:ident $fields:tt
        )*
    ) => {
        $(
            $(#[$meta])*
            $vis struct $name $fields

            impl Template for $name {
                const NAME: &'static str = stringify!($name);
                const TEXT: &'static str = $text;
            }
        )*
        const TEMPLATES: &'static [&'static dyn DynTemplate] = &[$(&PhantomData::<$name>,)*];
    };
}

#[derive(Debug, Serialize)]
pub struct SubscriptionLoggedOutTemplateProviders {
    pub pretty_name: String,
    pub url: String,
}

templates! {
    #[text = r#"<head><title>Subscription</title></head>
<body>
<h1>Register or Sign In:</h1>
As an anti-spam measure, you need to register using a 3rd-party account:
<ul>
{{- for provider in providers -}}
<li><a href="{provider.url}">{provider.pretty_name}</a></li>
{{- endfor -}}
</ul>
</body>"#]
    #[derive(Debug, Serialize)]
    pub struct SubscriptionLoggedOutTemplate {
        pub providers: Vec<SubscriptionLoggedOutTemplateProviders>,
    }
    #[text = r#"<head><title>Subscription</title></head>
<body>
<p>Signed in as {email}</p>
<ul>
<li><a href="/subscription/logout">Logout</a></li>
<li><b><a href="/subscription/unsubscribe">Unsubscribe and Delete Account</a></b></li>
</ul>
</body>"#]
    #[derive(Debug, Serialize)]
    pub struct SubscriptionLoggedInTemplate {
        pub email: EndUserEmail,
    }
}

pub fn make_templates() -> TinyTemplate<'static> {
    let mut templates = TinyTemplate::new();
    for template in TEMPLATES {
        if let Err(e) = templates.add_template(template.name(), template.text()) {
            panic!("error parsing template {}: {}", template.name(), e);
        }
    }
    templates
}

pub fn with_templates<F: FnOnce(&TinyTemplate<'static>) -> R, R>(f: F) -> R {
    thread_local! {
        static TEMPLATES: TinyTemplate<'static> = make_templates();
    }
    TEMPLATES.with(f)
}

#[derive(Debug, Clone, Deserialize, Serialize)]
pub struct SessionState {
    pub email: EndUserEmail,
}

impl SessionState {
    pub fn get(session: &Session) -> Option<Self> {
        session
            .get("state")
            .expect("state is signed so we know it's valid")
    }
    pub fn set(&self, session: &Session) {
        session
            .insert("state", self)
            .expect("serialization can't fail");
    }
}

pub async fn get_or_insert_account(
    email: EndUserEmail,
    session: &Session,
    db: web::Data<DbThread>,
) -> SessionState {
    let state = SessionState {
        email: email.clone(),
    };
    db.run_on(move |connection| {
        use crate::schema::accounts::dsl::accounts;
        use diesel::prelude::*;
        let account = Account {
            email: email.to_string(),
            date: chrono::Utc::now().naive_utc(),
        };
        match diesel::insert_into(accounts)
            .values(&account)
            .on_conflict_do_nothing()
            .execute(connection)
        {
            Ok(0) => {
                log::debug!("found existing account: {}", account.email.as_str());
            }
            Ok(_) => {
                log::info!("created new account: {}", account.email.as_str());
            }
            Err(e) => {
                log::error!(
                    "failed to get/insert account: {}\n{e}",
                    account.email.as_str()
                )
            }
        }
    })
    .await;
    state.set(session);
    state
}

const NONCE: &'static str = "nonce";
const CSRF_TOKEN: &'static str = "csrf_token";

#[derive(Deserialize)]
struct AuthQuery {
    code: AuthorizationCode,
    state: CsrfToken,
}

#[get("/subscription/callback/{provider}")]
pub async fn callback(
    config: web::Data<Config>,
    session: Session,
    path: web::Path<(String,)>,
    query: web::Query<AuthQuery>,
    db: web::Data<DbThread>,
) -> impl Responder {
    let mut resp = HttpResponse::SeeOther().body("");
    resp.headers_mut()
        .insert(LOCATION, "/subscription".parse().unwrap());
    if SessionState::get(&session).is_some() {
        return resp;
    }
    let body = async {
        let csrf_token = session
            .remove_as::<CsrfToken>(CSRF_TOKEN)
            .ok_or_eyre("missing csrf token")?
            .ok()
            .ok_or_eyre("invalid csrf token")?;
        let nonce = session
            .remove_as::<Nonce>(NONCE)
            .ok_or_eyre("missing nonce")?
            .ok()
            .ok_or_eyre("invalid nonce")?;
        ensure!(
            csrf_token.secret() == query.state.secret(),
            "csrf token doesn't match"
        );
        let provider = config.oidc.get(&path.0).ok_or_eyre("unknown provider")?;
        let token = provider
            .state()
            .client
            .exchange_code(query.into_inner().code)
            .request_async(async_http_client)
            .await?;
        if let Some(id_token) = token.id_token() {
            let id_token_claims =
                id_token.claims(&provider.state().client.id_token_verifier(), &nonce)?;
            if let Some(email) = id_token_claims.email().cloned() {
                get_or_insert_account(email, &session, db).await;
                return Ok(());
            }
        }
        let user_info: CoreUserInfoClaims = provider
            .state()
            .client
            .user_info(token.access_token().clone(), None)?
            .request_async(async_http_client)
            .await?;
        let email = user_info.email().ok_or_eyre("no email provided")?.clone();
        get_or_insert_account(email, &session, db).await;
        Ok(())
    };
    let result: eyre::Result<()> = body.await;
    if let Err(e) = result {
        log::warn!("callback error: {e}");
    }
    resp
}

#[get("/subscription/login/{provider}")]
pub async fn login(
    config: web::Data<Config>,
    session: Session,
    path: web::Path<(String,)>,
) -> impl Responder {
    let mut resp = HttpResponse::SeeOther().body("");
    resp.headers_mut()
        .insert(LOCATION, "/subscription".parse().unwrap());
    if SessionState::get(&session).is_some() {
        return resp;
    }
    let Some(provider) = config.oidc.get(&path.0) else {
        return resp;
    };
    let (url, csrf_token, nonce) = provider
        .state()
        .client
        .authorize_url(
            CoreAuthenticationFlow::AuthorizationCode,
            CsrfToken::new_random,
            Nonce::new_random,
        )
        .add_scopes(provider.scopes.clone())
        .url();
    session.insert(CSRF_TOKEN, csrf_token).unwrap();
    session.insert(NONCE, nonce).unwrap();
    resp.headers_mut()
        .insert(LOCATION, url.as_str().parse().unwrap());
    resp
}

#[get("/subscription/logout")]
pub async fn logout(session: Session) -> impl Responder {
    session.purge();
    let mut resp = HttpResponse::SeeOther().body("");
    resp.headers_mut()
        .insert(LOCATION, "/subscription".parse().unwrap());
    resp
}

#[get("/subscription/unsubscribe")]
pub async fn unsubscribe(session: Session, db: web::Data<DbThread>) -> impl Responder {
    if let Some(state) = SessionState::get(&session) {
        db.run_on(move |connection| {
            use crate::schema::accounts::dsl::*;
            use diesel::prelude::*;
            match diesel::delete(accounts.filter(email.eq(state.email.as_str())))
                .execute(connection)
            {
                Ok(_) => {
                    log::info!("deleted account: {}", state.email.as_str());
                }
                Err(e) => {
                    log::error!("failed to delete account: {}\n{e}", state.email.as_str())
                }
            }
        })
        .await;
    }
    session.purge();
    let mut resp = HttpResponse::SeeOther().body("");
    resp.headers_mut()
        .insert(LOCATION, "/subscription".parse().unwrap());
    resp
}

#[get("/subscription")]
pub async fn subscription(
    config: web::Data<Config>,
    session: Session,
    db: web::Data<DbThread>,
) -> impl Responder {
    if let Some(SessionState { email }) = SessionState::get(&session) {
        get_or_insert_account(email.clone(), &session, db).await;
        let template = SubscriptionLoggedInTemplate { email };
        HttpResponse::Ok()
            .content_type(ContentType::html())
            .body(template.render().expect("rendering can't fail"))
    } else {
        let mut template = SubscriptionLoggedOutTemplate { providers: vec![] };
        for (name, provider) in config.oidc.iter() {
            template
                .providers
                .push(SubscriptionLoggedOutTemplateProviders {
                    pretty_name: provider.pretty_name.clone(),
                    url: format!("/subscription/login/{name}"),
                });
        }
        HttpResponse::Ok()
            .content_type(ContentType::html())
            .body(template.render().expect("rendering can't fail"))
    }
}

pub fn all_services(cfg: &mut ServiceConfig) {
    cfg.service(subscription)
        .service(login)
        .service(callback)
        .service(logout)
        .service(unsubscribe);
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_make_templates() {
        make_templates();
    }
}
